#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include <pthread.h>
#include <errno.h>
#include <time.h>

#include "globals.h"
#include "gui.h"

#define LEFTBUTTON	1	/* Left mousebutton id */
#define MIDDLEBUTTON	2	/* Middle mousebutton id */
#define RIGHTBUTTON	3	/* Right mousebutton id */

#include "pixmaps/frame.xpm"
#include "pixmaps/newmail.xpm"
#include "pixmaps/oldmail.xpm"
#include "pixmaps/nomail.xpm"
#include "pixmaps/newmail_s.xpm"
#include "pixmaps/oldmail_s.xpm"
#include "pixmaps/nomail_s.xpm"
#define XPMCLOSENESS 32768
struct XpmIcon {
	Pixmap pixmap;
	Pixmap mask;
	XpmAttributes attributes;
	struct XpmIcon *next;
};
struct XpmIcon *Frame = NULL, 
	*NoMail = NULL, 
	*OldMail = NULL, 
	*NewMail = NULL, 
	*Current = NULL;
int saved_status = MAIL_NONE;
int need_animation = 0;

pthread_mutex_t x11_mutex = PTHREAD_MUTEX_INITIALIZER;
Widget MainWin;
GC WinGC;
XGCValues gcv;
XtAppContext AsmailApp;
Display *Disp = 0;
Window Root;
Atom wm_delete_win;
void Quit(Widget, XEvent *, String *, Cardinal *);
XtActionsRec asmail_actions[] = { { "Quit", Quit } };

XTextItem NumOfMsg;
char NumOfMsgText[MAX_INPUT_LENGTH+1];
Pixel NumOfMsgColor;
XFontStruct * font_struct;

Pixel GetColor(const char * ColorName) {
	XColor Color;
	XWindowAttributes Attributes;

	XGetWindowAttributes(Disp, Root, &Attributes);
	Color.pixel = 0;
	if (!XParseColor(Disp, Attributes.colormap, ColorName, &Color))
		printf("asmail: get_color: can't parse color %s\n", ColorName);
	else if (!XAllocColor(Disp, Attributes.colormap, &Color))
		printf("asmail: get_color: can't allocate color %s\n", ColorName);
	return Color.pixel;
}   

void XPMError(int Code) {
	switch (Code) {
	case 0:
		break;
	case 1:
	case -4:
		printf("asmail: not enough free color cells\n");
		break;
	case -1:
	case -2:
		printf("asmail: could not load xpm\n");
		break;
	case -3:
		printf("asmail: not enough memory free\n");
		break;
	default:
		printf("asmail: unknown xpm-error\n");
		break;
	}
	if (Code != 0)
		exit(1);
}

void LoadXPM(struct pixfile * filelist, struct XpmIcon ** destination) {
	struct XpmIcon * icon;
	struct pixfile * file_ptr;

	file_ptr = filelist;
	while ( file_ptr ) {
		icon = (struct XpmIcon *) calloc(1, sizeof(struct XpmIcon));
		icon->attributes.valuemask |= XpmCloseness;
		icon->attributes.closeness = XPMCLOSENESS;
		XPMError(XpmReadFileToPixmap(Disp, Root,
					file_ptr->name,
					&icon->pixmap,
					&icon->mask,
					&icon->attributes));
		icon->next = *destination;
		*destination = icon;
		file_ptr = file_ptr->next;
	}
}

struct XpmIcon *GetXPM(char **data) {
	struct XpmIcon *icon = 
		(struct XpmIcon *) calloc(1, sizeof(struct XpmIcon));

	icon->attributes.valuemask |= XpmCloseness;
	icon->attributes.closeness = XPMCLOSENESS;

	XPMError(XpmCreatePixmapFromData(Disp, Root,
					 data,
					 &icon->pixmap,
					 &icon->mask,
					 &icon->attributes));

	icon->next = (struct XpmIcon *) icon;

	return icon;
}

void execute_on_new() {
	if ( x11_set.beep )
		XBell(Disp, 0);
	if ( strlen(x11_set.on_new_mail) )
		system(x11_set.on_new_mail);
}

void add_status(char * line, int s) {
	if ( s & STAT_RUN )
		sprintf(&line[strlen(line)], "R");
	else
		sprintf(&line[strlen(line)], " ");
	if ( s & STAT_FAIL )
		sprintf(&line[strlen(line)], "F");
	else if ( s & STAT_CONN )
		sprintf(&line[strlen(line)], "C");
	else if ( s & STAT_LOGIN )
		sprintf(&line[strlen(line)], "L");
	else if ( s & STAT_TIMEOUT )
		sprintf(&line[strlen(line)], "T");
	else
		sprintf(&line[strlen(line)], " ");
}

int update_window() {
	struct mbox_struct * mb;
	int o = 0;
	int n = 0;
	int s = STAT_IDLE;
	int x = 0;
	int y = 0;
	XSizeHints SizeHints;
	int direction_return, ascent_return, descent_return;
	XCharStruct overall_return;
	int mail_status;

	mb = mbox;
	while (mb) {
		if ( mb->ctotal )
			o += mb->ctotal;
		if ( mb->cnew )
			n += mb->cnew;
		if ( mb->flags & FLAG_ARRIVED ) {
			pthread_mutex_lock(&mb->mutex);
			mb->flags &= ! FLAG_ARRIVED;
			pthread_mutex_unlock(&mb->mutex);
			execute_on_new();
		}
		s |= mb->status;
		mb = mb->next;
	}

	mail_status = n ? MAIL_NEW : o ? MAIL_OLD : MAIL_NONE;
	if ( (mail_status == saved_status) && (Current->next != NULL) ) {
		Current = Current->next;
	} else {
		if ( n )
			Current = NewMail;
		else if (o)
			Current = OldMail;
		else
			Current = NoMail;
	}
	saved_status = mail_status;
	if ( Current->next || 
			( (Current != NewMail) && 
			  (Current != OldMail) && 
			  (Current != NoMail) ) ) {
		need_animation = 1;
	} else {
		need_animation = 0;
	}

	if ( x11_set.shape ) {
		SizeHints.width = Current->attributes.width;
		SizeHints.height = Current->attributes.height;
		XtResizeWidget(MainWin, SizeHints.width, SizeHints.height, 1);
		SizeHints.flags = USPosition | PSize | PMinSize | PMaxSize;
		SizeHints.min_width = SizeHints.max_width = SizeHints.width;
		SizeHints.min_height = SizeHints.max_height = SizeHints.height;
		XSetWMNormalHints(Disp, XtWindow(MainWin), &SizeHints);
		XShapeCombineMask(Disp, XtWindow(MainWin), ShapeBounding, 0, 0,
				Current->mask, ShapeSet);
	} else {
		XCopyArea(Disp, Frame->pixmap, XtWindow(MainWin), WinGC, 0, 0,
				Frame->attributes.width, 
				Frame->attributes.height, 0, 0);
		x = (Frame->attributes.width - Current->attributes.width) / 2;
		y = (Frame->attributes.height - Current->attributes.height) / 2;
	}
	XCopyArea(Disp, Current->pixmap, XtWindow(MainWin), WinGC, 0, 0,
			Current->attributes.width, Current->attributes.height, 
			x, y);
	/*
	 * Display the summary of the mailboxes
	 */
	if ( x11_set.total ) {
		NumOfMsg.chars[0] = '\0';
		if ( x11_set.status ) {
			add_status(NumOfMsg.chars, s);
			sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)], " ");
		}
		if ( x11_set.new ) {
			if ( 0 != n )
				sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)],
						"%d", n);
			else
				sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)],
						"-");
		}
		if ( x11_set.old && x11_set.new ) {
			sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)], 
					x11_set.delimiter);
		}
		if ( x11_set.old ) {
			if ( 0 != o )
				sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)],
						"%d", o);
			else
				sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)],
						"-");
		}
		gcv.foreground = NumOfMsgColor;
		XChangeGC(Disp, WinGC, GCForeground, &gcv);
		NumOfMsg.nchars = strlen(NumOfMsg.chars);
		XDrawText(Disp, XtWindow(MainWin), WinGC, 
				x + x11_set.x, y + x11_set.y, 
				&NumOfMsg, 1);
		XTextExtents(font_struct, NumOfMsg.chars, 
				strlen(NumOfMsg.chars),
				&direction_return, &ascent_return,
				&descent_return, &overall_return);
		y += ascent_return + descent_return + 2;
	}
	/*
	 * Display information for each mailbox
	 */
	if ( x11_set.each ) {
		mb = mbox;
		while (mb) {
			NumOfMsg.chars[0] = '\0';
			if ( x11_set.status ) {
				add_status(NumOfMsg.chars, mb->status);
				sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)], " ");
			}
			if ( x11_set.new ) {
				if ( 0 != mb->cnew )
					sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)], "%d", mb->cnew);
				else
					sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)], "-");
			}
			if ( x11_set.old && x11_set.new ) {
				sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)],
						x11_set.delimiter);
			}
			if ( x11_set.old ) {
				if ( 0 != mb->ctotal )
					sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)], "%d", mb->ctotal);
				else
					sprintf(&NumOfMsg.chars[strlen(NumOfMsg.chars)], "-");
			}
			gcv.foreground = NumOfMsgColor;
			XChangeGC(Disp, WinGC, GCForeground, &gcv);
			NumOfMsg.nchars = strlen(NumOfMsg.chars);
			XDrawText(Disp, XtWindow(MainWin), WinGC, 
					x + x11_set.x, y + x11_set.y, 
					&NumOfMsg, 1);
			XTextExtents(font_struct, NumOfMsg.chars, 
					strlen(NumOfMsg.chars),
					&direction_return, &ascent_return,
					&descent_return, &overall_return);
			y += ascent_return + descent_return + 2;

			mb = mb->next;
		}
	}

	XFlush(Disp);

	return 0;
}

void ButtonHandler(Widget W, XtPointer P, XEvent * E) {
	if (E->xbutton.button == LEFTBUTTON) {
		if (strlen(x11_set.on_left))
			system(x11_set.on_left);
	} else if (E->xbutton.button == MIDDLEBUTTON) {
		if (strlen(x11_set.on_middle))
			system(x11_set.on_middle);
	} else if (E->xbutton.button == RIGHTBUTTON) {
		if (strlen(x11_set.on_right))
			system(x11_set.on_right);
	}
}

void ExposeHandler(Widget W, XtPointer P, XEvent * E) {
	signal_xkick();
}

void StructureNotifyHandler(Widget W, XtPointer P, XEvent * E) {
	if ( MapNotify == E->type ) {
		signal_xkick();
	}
}

void Quit(Widget w, XEvent *event, String *params, Cardinal *num_params) {
	if (event->type == ClientMessage &&
			event->xclient.data.l[0] != wm_delete_win) {
		XBell(Disp, 0);
		return;
	}
	XFreeGC(Disp, WinGC);
	XCloseDisplay(Disp);
	exit (0);
}

void main_x_loop(void * ptr) {
	XtAppMainLoop(AsmailApp);
	exit(0);
}

void animation_kicker(void * ptr) {
	struct timespec tm;
	tm.tv_sec = x11_set.refresh / 100;
	tm.tv_nsec = x11_set.refresh % 100 * 10000000;
	while (1) {
		nanosleep(&tm, NULL);
		if ( need_animation )
			signal_xkick();
	}
}

void startx(void * ptr) {
	XSizeHints SizeHints;
	unsigned long gcm;
	int kicks;

	int set_x,set_y,set_w,set_h;

	if ( pthread_mutex_trylock(&x11_mutex) == EBUSY ) {
		printf("asmail: startx: non-reentrant function!\n");
		exit(0);
	}

	MainWin = XtAppInitialize(&AsmailApp, x11_set.title, NULL, 0, 
			&x11_set.argc, x11_set.argv, NULL, NULL, 0);
	Disp = XtDisplay(MainWin);
	Root = RootWindow(Disp, DefaultScreen(Disp));

	/* Here we have 2 possibilities. Either there are no pics
	 * given in the config file -> use the built-in and no
	 * animation. Or we got a list of pics and we will have
	 * to animate through that list. */
	if ( x11_set.nomail ) {
		LoadXPM(x11_set.nomail, &NoMail);
	} else {
		NoMail = GetXPM((x11_set.shape) ? nomail_s : nomail);
		NoMail->next = NoMail;
	}
	if ( x11_set.oldmail ) {
		LoadXPM(x11_set.oldmail, &OldMail);
	} else {
		OldMail = GetXPM((x11_set.shape) ? oldmail_s : oldmail);
		OldMail->next = OldMail;
	}
	if ( x11_set.newmail ) {
		LoadXPM(x11_set.newmail, &NewMail);
	} else {
		NewMail = GetXPM((x11_set.shape) ? newmail_s : newmail);
		NewMail->next = NewMail;
	}
	Frame = GetXPM(frame);
	Current = NoMail;

	SizeHints.width = Frame->attributes.width;
	SizeHints.height = Frame->attributes.height;
	if ( strlen(x11_set.geometry) ) {
		XParseGeometry( x11_set.geometry,
				&SizeHints.x, &SizeHints.y,
				&SizeHints.width, &SizeHints.height );
	}
	XtResizeWidget(MainWin, SizeHints.width, SizeHints.height, 1);
	XtRealizeWidget(MainWin);
	SizeHints.flags = USPosition | PSize | PMinSize | PMaxSize;
	SizeHints.min_width = SizeHints.max_width = SizeHints.width;
	SizeHints.min_height = SizeHints.max_height = SizeHints.height;
	XSetWMNormalHints(Disp, XtWindow(MainWin), &SizeHints);

	gcm = GCForeground | GCBackground | GCGraphicsExposures;
	gcv.foreground = GetColor("white");
	gcv.background = GetColor("white");
	gcv.graphics_exposures = False;
	WinGC = XCreateGC(Disp, Root, gcm, &gcv);

	XtAddEventHandler(MainWin, ExposureMask, FALSE, (XtEventHandler) ExposeHandler, NULL);
	XtAddEventHandler(MainWin, ButtonPressMask, FALSE, (XtEventHandler) ButtonHandler, NULL);
	XtAddEventHandler(MainWin, StructureNotifyMask, FALSE, (XtEventHandler) StructureNotifyHandler, NULL);
	XtAppAddActions(AsmailApp, asmail_actions, XtNumber(asmail_actions));
	XtOverrideTranslations(MainWin,
		XtParseTranslationTable("<Message>WM_PROTOCOLS: Quit()"));
	wm_delete_win = XInternAtom(Disp, "WM_DELETE_WINDOW", False);
	(void) XSetWMProtocols(Disp, XtWindow(MainWin), &wm_delete_win, 1);

	XtMapWidget(MainWin);

	NumOfMsg.chars = NumOfMsgText;
	NumOfMsg.font = XLoadFont(Disp, x11_set.font);
	NumOfMsgColor = GetColor(x11_set.color);
	font_struct = XQueryFont(Disp, NumOfMsg.font);

	pthread_create( &x11_set.main_x_thread, pthread_attr_default,
		(void*)&main_x_loop, (void*) &x11_set);
	pthread_detach( x11_set.main_x_thread );

	if ( x11_set.nomail || x11_set.oldmail || x11_set.newmail ) {
		pthread_create( &x11_set.animation_thread, pthread_attr_default,
			(void*)&animation_kicker, (void*) &x11_set);
		pthread_detach( x11_set.animation_thread );
	}

	while ( 1 ) {
		kicks = sleep_xkick();
		update_window();
	}

	XFreeGC(Disp, WinGC);
	XCloseDisplay(Disp);
	pthread_exit(NULL);
}

